from pathlib import Path

# Filesystem
ROOT = Path().absolute().parent
RESC = ROOT / "resc"
assert RESC.exists()
ELKI_FILEPATH = RESC / "elki.jar"
ELKI_URL = "https://elki-project.github.io/releases/release0.7.5/elki-bundle-0.7.5.jar"



from .Elki import Elki
from prefect.tasks.shell import ShellTask
from prefect import Flow
import pandas as pd
import io

from .cte import ELKI_FILEPATH


class Hics(Elki):
    def __init__(
        self,
        verbose=False,
        elki=ELKI_FILEPATH,
        contamination=0.1,
        k=5,
        **kwargs
    ):
        super().__init__(verbose=verbose, elki=elki, contamination=contamination)

        self.k = k
        self.fit_data_fp = None

        return

    def fit(self, X):
        super().fit(X)

        # Run
        self.status = self.fit_flow.run()
        self.out = self.status.result[self.fit_shell_task].result

        # Process
        df = self._to_dataframe(self.out)

        # Set scores and labels
        # TODO: Make this more consistent.
        self._scores = self._to_scores(df)
        self._set_labels()

        return

    # --------------
    # Internal Methods
    # --------------
    @staticmethod
    def _to_dataframe(out):
        return pd.read_csv(
            io.StringIO("\n".join(out)), delim_whitespace=True, header=None, index_col=0
        )

    @staticmethod
    def _to_scores(dataframe):
        return dataframe.sort_index().values.squeeze()

    # --------------
    # CLI Properties
    # --------------
    @property
    def fit_flow(self):
        with Flow("fit") as f:
            self.fit_shell_task = ShellTask(return_all=True, log_stderr=True)(
                command=self.fit_command
            )

        return f

    @property
    def fit_command(self):
        fit_cmd = "{} {} {} {} {}".format(
            self.kdd_command,
            self.i_command,
            self.hics_command,
            self.eval_command,
            self.o_command,
        )
        return fit_cmd

    @property
    def hics_command(self):
        return "-algorithm outlier.meta.HiCS -lof.k {}".format(self.k)

    @property
    def i_command(self):
        # Command snippet for input data
        return "-db HashmapDatabase -dbc.in {}".format(self.data_filepath.name)



import numpy as np
import pandas as pd
import prefect

import tempfile

from .cte import ELKI_FILEPATH


class Elki:
    def __init__(self, verbose=False, elki=ELKI_FILEPATH, contamination=0.1, **kwargs):
        # General params
        self.verbose = verbose

        # Metadata
        self.n_instances = None
        self.n_attributes = None

        # Data
        self.data_filepath = None

        # Outcomes
        self.contamination = contamination
        self._scores = None
        self._labels = None

        # CLI
        self.elki = elki
        return

    def fit(self, X):
        self._init_metadata(X)
        self._init_data(X)
        self._init_labels()
        return

    def predict(self):
        return self.labels

    # --------------
    # Internal Methods
    # --------------
    def _init_metadata(self, X):
        self.n_instances, self.n_attributes = X.shape
        return

    def _init_data(self, data):
        filepath = tempfile.NamedTemporaryFile(suffix=".csv")
        np.savetxt(filepath.name, data, delimiter=",")
        self.data_filepath = filepath
        return

    def _init_labels(self):
        self._labels = np.zeros(self.n_instances)
        return

    def _set_labels(self):
        anomaly_idxs = np.argpartition(self.scores, -self.integer_contamination)[
            -self.integer_contamination :
        ]
        self._labels[anomaly_idxs] = 1
        return

    # --------------
    # Properties
    # --------------
    @property
    def contamination(self):
        return self._contamination

    @contamination.setter
    def contamination(self, value):
        assert (
            0.0 <= value <= 1.0
        ), "Contamination must be a percentage, expressed as a float in [0.0, 1.0]."
        self._contamination = value
        return

    @property
    def decision_scores_(self):
        return self.scores

    @property
    def scores(self):
        return self._scores

    @property
    def labels(self):
        return self._labels

    @labels.setter
    def labels(self, value):
        assert isinstance(value, np.ndarray)
        assert value.shape[0] == self.n_instances
        self._labels == value
        return

    @property
    def n_anomalies(self):
        return np.sum(self.labels)

    @property
    def integer_contamination(self):
        # Contamination in integer.
        return int(self.contamination * self.n_instances)

    # CLI properties
    @property
    def kdd_command(self):
        kddcli = "KDDCLIApplication"
        return "{} {}".format(self.javacall, kddcli)

    @property
    def javacall(self):
        return "java -jar {}".format(self.elki)

    @property
    def eval_command(self):
        # Command snippet for evaluation
        return "-evaluator NoAutomaticEvaluation"

    @property
    def o_command(self):
        # Command snippet for output data
        return "-resulthandler tutorial.outlier.SimpleScoreDumper"

    # --------------
    # Static Methods
    # --------------



# -*- coding: utf-8 -*-
from pkg_resources import get_distribution, DistributionNotFound
from requests import get  # to make GET request


def _download_elki(
    url=None,
    filepath=None,
):

    # open in binary mode
    with open(filepath, "wb") as file:
        # get request
        response = get(url)
        # write to file
        file.write(response.content)
    return


try:
    # Change here if project is renamed and does not equal the package name
    dist_name = __name__
    __version__ = get_distribution(dist_name).version

    from .cte import ELKI_FILEPATH, ELKI_URL

    # download elki jar, if necessary
    if not ELKI_FILEPATH.exists():
        _download_elki(
            url=ELKI_URL,
            filepath=ELKI_FILEPATH,
        )

    # imports
    from .Hics import Hics

except DistributionNotFound:
    __version__ = "unknown"
finally:
    del get_distribution, DistributionNotFound
